# =============================================================================
#　skBLEtool.py
# ブルートゥースのBLEデバイスと通信を行う為のツールです。
# このスクリプトには他に、skBLEsub.py/skDICT.pyが必要です。
# =============================================================================
# VERSION  DATE        BY                   CHANGE/COMMENT
# -----------------------------------------------------------------------------
# 1.00     2021-05-07  きむ茶工房(きむしげ)  Create
# 1.10     2021-06-22  きむ茶工房(きむしげ)  notify/readの表示型式変更
# -----------------------------------------------------------------------------
# IDE : PyCharm(COMMUNITY 2019.2) Python 3.7
# wxPython : 4.0.6　 bleak :
# Operating device : ESP32-WROOM-32E RN4020
# PDS(Public Domain Software)
# =============================================================================
import sys
import asyncio
import wx
import bleak
from bleak import BleakClient
from bleak import BleakScanner
from skBLEsub import win_mon

class win_app():
    # 変数の初期化
    TASK_CONSOLE_FLAG = False     # コンソール画面の起動状態フラグ
    TASK_SETGATT_FLAG = False     # GATT設定情報画面の起動状態フラグ
    TASK_RECORD_FLAG  = False     # 記録帳画面の起動状態フラグ
    CONSOLE_OBJ  = 0              # コンソール画面のインスタンス
    SERVICE_UUID = ''             # 選択されたサービスのUUID
    NOTIFY_TIMER = 10             # 通知待ちの有効期間
    record_fname = 'BLEtool'      # 記録帳のファイル名(アドレス,UUID等の保存用)
    scan_queue = asyncio.Queue()  # スキャン時のキューバッファ
    Countdown = 0                 # カウントダウン変数
    CountdownSetT = 6             # 6*10s(1分)でカウントダウンは終了(デフォルト)

    ###############################################################################
    # イベントハンドラ関数
    ###############################################################################
    #---- タイマーのイベント ----
    def OnTimer(self, event):
        """ 10秒で１回イベント発生　"""
        if self.Countdown == 0:
            wx.Bell()  # システムのベルを鳴らす。
            print("スキャンします。")
            _r = self.menu_scanner()  # スキャンの処理実行
            self.timer.Stop()         # timer停止
            if _r is None:
                print('デバイスが存在しません。')
            else:
                print(f'{_r} が見つかりました。')
                _kosu = len(win_mon.CHARACTERISTIC_UUID)
                if _kosu == 0:
                    print('キャラクタリスティックUUIDが選択されていません。')
                else:
                    self.menu_read()  # readの処理実行
            print('[Scan + read]機能を再開します。')
            self.Countdown = self.CountdownSetT
            self.timer.Start(10000)   # timer起動、10秒間隔で再起動
        else:
            self.Countdown -= 1

    #---- クローズ時のイベント ----
    def FrameClose(self, event):
        """ メイン画面のクローズボタンが押された時の処理 """
        import gc
        self.timer.Stop()  # 停止させていないかも取りあえずtimer停止
        gc.collect()
        wx.Exit()          # メインウインドウ終了

    def mon_winClose(self, event):
        """ コンソール画面のクローズボタンが押された時の処理 """
        _obj = event.GetEventObject()
        self.TASK_CONSOLE_FLAG = False
        sys.stdout = sys.__stdout__  # 標準出力に戻す
        sys.stderr = sys.__stderr__  # 標準エラー出力に戻す
        _obj.Destroy()               # コンソールウインドウ終了

    def setgatt_winClose(self, event):
        """ GATT設定情報画面のクローズボタンが押された時の処理 """
        _obj = event.GetEventObject()
        self.TASK_SETGATT_FLAG = False
        _obj.Destroy()       # コンソールウインドウ終了

    def record_winClose(self, event):
        """ 記録帳画面のクローズボタンが押された時の処理 """
        _obj = event.GetEventObject()
        self.TASK_RECORD_FLAG = False
        _obj.Destroy()       # コンソールウインドウ終了

    #---- メニュー選択時のイベント ----
    def SelectSubMenu(self, event):
        """ 記録帳リストで右クリックのサブメニュ項目が選択された時の処理 """
        _obj_id = event.GetId()      # 項目のIDを得る
        if _obj_id == 100:
            # 【読込む】
            self.sub_menu_reads()
        elif _obj_id == 101:
            # 【削除】
            self.sub_menu_deletes()

    def SelectMenu(self, event):
        """ メニュー項目が選択された時の処理 """
        _obj_id = event.GetId()      # 項目のIDを得る
        if _obj_id == wx.ID_EXIT:
            # 【終了】
            # メッセージ表示ダイアログを作成
            msg_dlg = wx.MessageDialog(None,
                                       '終了します。よろしいですか？', '終了確認',
                                       wx.YES_NO | wx.ICON_INFORMATION)
            _ans = msg_dlg.ShowModal()
            msg_dlg.Destroy()
            if _ans == wx.ID_YES:
                self.main_win.Close()
        elif _obj_id == 1:
            # 【検出】
            self.menu_detection()
        elif _obj_id == 2:
            # 【スキャン】
            self.menu_scanner()
        elif _obj_id == 3:
            # 【サービス】
            self.menu_service()
        elif _obj_id == 4:
            # 【キャラクタリスティック】
            self.menu_char()
        elif _obj_id == 5:
            # 【通知】
            self.menu_notify()
        elif _obj_id == 6:
            # 【コンソール】
            self.menu_console()
        elif _obj_id == 7:
            # 【GATT情報】
            self.menu_setgatt()
        elif _obj_id == 8:
            # 【read】
            self.menu_read()
        elif _obj_id == 9:
            # 【オブザーバー】
            self.menu_observer()
        elif _obj_id == 10:
            # 【"Scan + read"の開始】
            self.sub_menu_start_func1()
        elif _obj_id == 11:
            # 【"Scan + read"の停止】
            self.timer.Stop()        # timer停止
            print('[Scan + read]機能を停止しました。')

        elif _obj_id == wx.ID_HELP:
            # 【操作】
            import webbrowser
            webbrowser.open('http://zattouka.net/GarageHouse/soft/BLEtool/BLEtool.htm#operation')      # マニュアルのページを開く

    #---- リスト選択時のイベント ----
    def detect_list_select(self, event):
        """ 検出リスト選択時のイベントハンドラー """
        _obj = event.GetEventObject()
        # 選択したリストのインデックスを取得する
        _index = _obj.GetFirstSelected()
        # インデックスのリストからアドレスを取得する
        win_mon.DEVICE_ADDRESS = _obj.GetItemText(_index, 1)       # アドレスの項目

    def detect_list_deselect(self, event):
        """ 検出リスト選択を解除された時のイベントハンドラー """
        win_mon.DEVICE_ADDRESS = ""

    def service_list_select(self, event):
        """ サービスリスト選択時のイベントハンドラー """
        _obj = event.GetEventObject()
        # 選択したリストのインデックスを取得する
        _index = _obj.GetFirstSelected()
        # インデックスのリストからアドレスを取得する
        self.SERVICE_UUID = _obj.GetItemText(_index, 0)            # UUIDの項目

    def service_list_deselect(self, event):
        """ サービスリスト選択を解除された時のイベントハンドラー """
        self.SERVICE_UUID = ""

    def char_list_select(self, event):
        """ キャラクタリスティックスト選択時のイベントハンドラー """
        win_mon.CHARACTERISTIC_ATTR.clear()
        win_mon.CHARACTERISTIC_UUID.clear()        # 全てクリア
        _obj = event.GetEventObject()
        _n = _obj.GetSelectedItemCount()           # 選択個数
        _index = _obj.GetFirstSelected()           # 最初の選択項目
        win_mon.CHARACTERISTIC_UUID.append(_obj.GetItemText(_index, 1))      # 最初のUUIDの項目セット
        _s = _obj.GetItemText(_index, 0) + ':' + _obj.GetItemText(_index, 2)
        win_mon.CHARACTERISTIC_ATTR.append(_s)                                  # 最初のtype/handleの項目セット
        for _i in range(0, _n-1):
            _index = _obj.GetNextSelected(_index)  # 次の選択項目
            win_mon.CHARACTERISTIC_UUID.append(_obj.GetItemText(_index, 1))  # 次のUUIDの項目セット
            _s = _obj.GetItemText(_index, 0) + ':' + _obj.GetItemText(_index, 2)
            win_mon.CHARACTERISTIC_ATTR.append(_s)                              # 次のtype/handleの項目セット
        self.main_win.SetStatusText('≪　通信の準備が整いました　≫')

    def char_list_deselect(self, event):
        """ キャラクタリスティックリスト選択を解除された時のイベントハンドラー """
        win_mon.CHARACTERISTIC_ATTR.clear()
        win_mon.CHARACTERISTIC_UUID.clear()

    #---- ボタンを押された時のイベント ----
    def click_clear_btn(self, event):
        """ コンソール画面のクリアボタンを押した時のイベントハンドラー """
        self.CONSOLE_OBJ.cls()

    def notify_textctrl_enter(self, event):
        """ 通知待ちの有効期間テキスト入力時にEnterキーを押された時のイベントハンドラー """
        _obj = event.GetEventObject()
        _n = _obj.GetValue()
        if _n.isnumeric():
            self.NOTIFY_TIMER = _n
            self.look_btn.SetFocus()  # Enterキーを押したと解かる様にフォーカスを移す
            self.main_win.SetStatusText('')
        else:
            self.main_win.SetStatusText('正しく入力されていない様です。')

    def adrs_textctrl_enter(self, event):
        """ デバイスアドレスのテキスト入力時にEnterキーを押された時のイベントハンドラー """
        _obj = event.GetEventObject()
        win_mon.DEVICE_ADDRESS = _obj.GetValue()
        self.look_btn.SetFocus()  # Enterキーを押したと解かる様にフォーカスを移す
        # デバイスのアドレスリストに表示する
        self.detect_list.DeleteAllItems()
        self.detect_list.InsertItem(0, "")
        self.detect_list.SetItem(0, 1, win_mon.DEVICE_ADDRESS)

    def uuid1_textctrl_enter(self, event):
        """ サービスUUIDのテキスト入力時にEnterキーを押された時のイベントハンドラー """
        _obj = event.GetEventObject()
        self.SERVICE_UUID = _obj.GetValue()
        self.look_btn.SetFocus()  # Enterキーを押したと解かる様にフォーカスを移す
        # サービスリストに表示する
        self.service_list.DeleteAllItems()
        self.service_list.InsertItem(0, self.SERVICE_UUID)

    def uuid2_textctrl_enter(self, event):
        """ キャラクタリスティックUUIDのテキスト入力時にEnterキーを押された時のイベントハンドラー """
        win_mon.CHARACTERISTIC_ATTR.clear()
        win_mon.CHARACTERISTIC_UUID.clear()  # 全てクリア
        _obj = event.GetEventObject()
        _s = _obj.GetValue().split(';')
        if len(_s) == 3:
            _s1 = _s[0] + ':' + _s[2]
            win_mon.CHARACTERISTIC_ATTR.append(_s1)
            win_mon.CHARACTERISTIC_UUID.append(_s[1])
            self.look_btn.SetFocus()  # Enterキーを押したと解かる様にフォーカスを移す
            self.char_list.DeleteAllItems()
            self.char_list.InsertItem(0, _s[0])
            self.char_list.SetItem(0, 1, _s[1])
            self.char_list.SetItem(0, 2, _s[2])
            _s = 'キャラクタリスティックのUUID：' + '[1/1件です。]'
            self.uuid2_come.SetLabel(_s)
            # スピンボタンのMAX値を設定する
            _kosu = len(win_mon.CHARACTERISTIC_UUID)
            if _kosu == 0:
                self.spin_button.SetMax(0)
            else:
                self.spin_button.SetMax(_kosu - 1)
            self.main_win.SetStatusText('')
        else:
            self.main_win.SetStatusText('正しく入力されていない様です。')

    def click_look_btn(self, event):
        """ 記録帳を見るボタンが押された時のイベントハンドラー """
        # 記録帳の画面作成を行う
        if not self.record_win_make():
            return
        # 記録のデータ表示を行う
        self.dict_display()
        # 記録帳フフレームの表示
        self.record_win.Show()

    def click_record_btn(self, event):
        """ 記録帳に登録ボタンが押された時のイベントハンドラー """
        # 記録帳の画面作成を行う
        if not self.record_win_make():
            return
        # 記録のデータ表示を行う
        self.dict_display()
        # 記録帳フフレームの表示
        self.record_win.Show()

        # キー値の入力処理
        _text_dlg = wx.TextEntryDialog(None, '記録帳ファイルへ登録するキーの名前を入力して下さい', 'キー値入力')
        _ans = _text_dlg.ShowModal()
        if _ans == wx.ID_OK:
            _kdt = _text_dlg.GetValue()  # 入力値を得る
            if _kdt:
                _text_dlg.Destroy()
                # データ値の作成
                _s = ''
                _i = 0
                _kosu = len(win_mon.CHARACTERISTIC_UUID)
                for _id in win_mon.CHARACTERISTIC_UUID:
                    _s = _s + _id + '|' + win_mon.CHARACTERISTIC_ATTR[_i] + ';'
                    _i = _i + 1
                _vdt = win_mon.DEVICE_ADDRESS + ';' + self.SERVICE_UUID + ';' + str(_kosu) + ';' + _s
                # 記録データに登録
                from skDICT import DICT
                _r = DICT()
                _r.open(self.record_fname)
                _r.put(_kdt, _vdt)
                _r.close()
                # 記録帳リストを空にする
                self.record_list.DeleteAllItems()
                # 記録帳のデータ表示を行う
                self.dict_display()

    def spin_up_down(self, event):
        """ 記録帳でスピンボタンの上下が押された時のイベントハンドラー """
        _obj = event.GetEventObject()
        _kosu = len(win_mon.CHARACTERISTIC_UUID)
        _n = int(_obj.GetValue())
        if _kosu != 0:
            self.uuid2_textctrl.SetValue(win_mon.CHARACTERISTIC_UUID[_n])
            _s = 'キャラクタリスティックのUUID：' + f'[{_n+1}/{_kosu}件です。]'
            self.uuid2_come.SetLabel(_s)

    #---- クリックされた時のイベント ----
    def record_right_click(self, event):
        """ 記録帳で右クリックされた時のイベントハンドラー """
        # サブメニュを作成する
        _p = wx.Panel(self.record_list, -1)
        menu = wx.Menu()
        menu.Append(100, "設定情報へ読込む")
        menu.Append(101, "記録帳から削除")
        menu.Bind(wx.EVT_MENU, self.SelectSubMenu)

        _p.PopupMenu(menu)
        menu.Destroy()

    ###############################################################################
    # 関数
    ###############################################################################
    #--- 機能メニュー関連の関数 ---
    def sub_menu_start_func1(self):
        """ "Scan + read"の開始 """
        if win_mon.DEVICE_ADDRESS != "":
            # キー値の入力処理
            _text_dlg = wx.TextEntryDialog(None, 'スキャンする間隔の時間(10秒単位)を入力して下さい', 'タイマー値入力')
            _text_dlg.SetValue(str(self.CountdownSetT))  # 現在の入力値
            _ans = _text_dlg.ShowModal()
            if _ans == wx.ID_OK:
                _kdt = _text_dlg.GetValue()              # 入力値を得る
                self.Countdown = self.CountdownSetT = int(_kdt)
                print('[Scan + read]機能を開始します。')
                self.timer.Start(10000)  # timer起動、10秒間隔で起動
        else:
            print('デバイスが選択されていません。')

    #---- 記録帳関連の関数 ----
    def sub_menu_reads(self):
        """ 記録帳リストから選択している項目を読み出す """
        _p = self.record_list.GetFocusedItem()
        _vdt = self.record_list.GetItem(_p, 1).GetText()  # データ値を得る
        _s = _vdt.split(';')
        # デバイスのアドレスをセット
        win_mon.DEVICE_ADDRESS = _s[0]
        self.adrs_textctrl.SetValue(_s[0])
        # サービスのUUIDをセット
        self.SERVICE_UUID = _s[1]
        self.uuid1_textctrl.SetValue(_s[1])
        # キャラクタリスティックのUUIDをセット
        win_mon.CHARACTERISTIC_ATTR.clear()
        win_mon.CHARACTERISTIC_UUID.clear()
        _kosu = int(_s[2])
        for _i in range(3, _kosu+3):
            _s1 = _s[_i].split('|')
            win_mon.CHARACTERISTIC_UUID.append(_s1[0])    # 登録数だけ読み出す
            win_mon.CHARACTERISTIC_ATTR.append(_s1[1])
        self.uuid2_textctrl.SetValue(win_mon.CHARACTERISTIC_UUID[0])  # 1個目をセット
        # デバイスのアドレスリストに表示する
        self.detect_list.DeleteAllItems()
        self.detect_list.InsertItem(0, "")
        self.detect_list.SetItem(0, 1, _s[0])
        _vdt = self.record_list.GetItem(_p, 0).GetText()  # 名前を得る
        self.detect_list.SetItem(0, 2, _vdt)
        self.detect_list.SetItemState(0, wx.LIST_STATE_SELECTED, wx.LIST_STATE_SELECTED)  # 選択状態にする
        # サービスリストに表示する
        self.service_list.DeleteAllItems()
        self.service_list.InsertItem(0, _s[1])
        # キャラクタリスティックリストに表示する
        self.char_list.DeleteAllItems()
        for _i in range(0, _kosu):
            _s1 = win_mon.CHARACTERISTIC_ATTR[_i].split(':')
            self.char_list.InsertItem(_i, _s1[0])
            self.char_list.SetItem(_i, 2, _s1[1])
            self.char_list.SetItem(_i, 1, win_mon.CHARACTERISTIC_UUID[_i])
        _s = 'キャラクタリスティックのUUID：' + f'[1/{_kosu}件です。]'
        self.uuid2_come.SetLabel(_s)
        # スピンボタンのMAX値を設定する
        _kosu = len(win_mon.CHARACTERISTIC_UUID)
        if _kosu == 0:
            self.spin_button.SetMax(0)
        else :
            self.spin_button.SetMax(_kosu-1)
        self.main_win.SetStatusText('≪　通信の準備が整いました　≫')

    def sub_menu_deletes(self):
        """ 記録帳リストから選択している項目を削除する """
        from skDICT import DICT
        _p = self.record_list.GetFocusedItem()
        _kdt = self.record_list.GetItem(_p, 0).GetText()  # キー値を得る
        # 記録データから削除する
        _r = DICT()
        _r.open(self.record_fname)
        _r.delete(_kdt)
        _r.close()
        # 記録帳リストを空にする
        self.record_list.DeleteAllItems()
        # 記録のデータ表示を行う
        self.dict_display()

    def dict_display(self):
        """ 記録帳リストに記録データ表示を行う """
        from skDICT import DICT
        _r = DICT()
        _r.open(self.record_fname)
        # キー値の表示
        _kdt = _r.keys()
        if len(_kdt) != 0:
            i = 0
            for k in _kdt:
                self.record_list.InsertItem(i, k)
                i = i + 1
        # データ値の表示
        _vdt = _r.values()
        if len(_vdt) != 0:
            i = 0
            for v in _vdt:
                self.record_list.SetItem(i, 1, v)
                i = i + 1
        _r.close()

    def record_win_make(self):
        """ 記録帳の画面作成 """
        self.main_win.SetStatusText('')
        if self.TASK_RECORD_FLAG:
            self.main_win.SetStatusText('既に起動中です。')
            return False
        self.TASK_RECORD_FLAG = True
        # [記録帳フレームを作成]
        self.record_win = wx.Frame(self.main_win, -1, '記録帳 - 項目を右クリックで[読込み][削除]の操作ができます。',
                                   size=(600, 300), style=wx.RESIZE_BORDER | wx.CLOSE_BOX | wx.CAPTION)
        self.record_win.SetBackgroundColour("LIGHT BLUE")
        self.record_win.Bind(wx.EVT_CLOSE, self.record_winClose)  # 閉じる[X]のイベント
        self.record_win.SetIcon(self.icon)                        # アイコンの設定

        # [記録帳リストコントロールの作成]
        self.record_list = wx.ListCtrl(self.record_win, wx.ID_ANY,
                                       style=wx.LC_AUTOARRANGE | wx.LC_REPORT, size=(600, 300))
        # カラム
        self.record_list.InsertColumn(0, "キー値", wx.LC_ALIGN_LEFT, 100)
        self.record_list.InsertColumn(1, "内容 (デバイスアドレス;サービスUUID;件数;キャラクタリスティックUUID|type:handle;...)", wx.LC_ALIGN_LEFT, 500)
        # イベントの登録
        self.record_list.Bind(wx.EVT_LIST_ITEM_RIGHT_CLICK, self.record_right_click)
        return True

    #---- オブザーバー関連関数 ----
    def menu_observer(self):
        """ ブロードキャスターデバイスからデータを得る処理 """
        _ldt = []

        async def run():
            _device = await BleakScanner.discover()
            for _d in _device:
                if _d.address == win_mon.DEVICE_ADDRESS:
                    _s = str(_d.metadata['manufacturer_data'])
                    _ldt.append(_d)

        if win_mon.DEVICE_ADDRESS == "":
            self.main_win.SetStatusText('デバイスが選択されていません。')
            return
        self.main_win.SetStatusText('【オブザーバー検出中】')
        # BLE通信タスクの起動を行う
        loop = asyncio.get_event_loop()
        loop.run_until_complete(run())
        # run()が終了するまでここは通過しない
        self.main_win.SetStatusText('')
        for _d in _ldt:
            _s = str(_d.metadata['manufacturer_data']).split(':')
            if len(_s) == 2:
                print(_s[1][:-1])   # データを表示
            else:
                print(f'{win_mon.DEVICE_ADDRESS}は、ブロードキャスターデバイスでない様です。')

    #---- デバイスへ通信を行う関連関数 ----
    def menu_detection(self):
        """ 周辺デバイスの検出を行う処理　"""
        _ldt = []

        async def run():
            async with BleakScanner() as scanner:
                await asyncio.sleep(5.0)  # 5秒間スキャン
                _devices = await scanner.get_discovered_devices()
                for d in _devices:
                    _ldt.append(d)

        self.main_win.SetStatusText('【検出中】')
        # BLE通信タスクの起動を行う
        loop = asyncio.get_event_loop()
        loop.run_until_complete(run())
        # run()が終了するまでここは通過しない
        self.detect_list.DeleteAllItems()   # 検出リストを空にする
        self.service_list.DeleteAllItems()  # サービスリストを空にする
        self.char_list.DeleteAllItems()     # キャラクタリスティックリストを空にする
        # 検出リストに追加する
        i = 0
        for l in _ldt:
            self.detect_list.InsertItem(i, '')
            self.detect_list.SetItem(i, 1, l.address)
            self.detect_list.SetItem(i, 2, l.name)
            self.detect_list.SetItem(i, 3, str(l.metadata['uuids']))
            self.detect_list.SetItem(i, 4, str(l.rssi))
            self.detect_list.SetItem(i, 5, str(l.details))
            self.detect_list.SetItem(i, 6, str(l.metadata['manufacturer_data']))
            i = i + 1
        if i == 0:
            self.main_win.SetStatusText('デバイスは検出されませんでした(Detected again)。')
        else:
            self.main_win.SetStatusText(f'デバイスを {i}件 検出しました。')

    def menu_scanner(self):
        """　選択されたデバイスをスキャンする処理　"""
        async def run():
            _device = None
            try:
                _device = await BleakScanner.find_device_by_address(win_mon.DEVICE_ADDRESS)
            except bleak.exc.BleakError as _e:
                print(_e)
            self.scan_queue.put_nowait(_device)  # キューバッファに追加

        if win_mon.DEVICE_ADDRESS == "":
            self.main_win.SetStatusText('デバイスが選択されていません。')
            return
        try:
            self.scan_queue.get_nowait()         # キューバッファを空にする
        except asyncio.QueueEmpty:
            pass
        self.main_win.SetStatusText('【スキャン中】')
        # BLE通信タスクの起動を行う
        loop = asyncio.get_event_loop()
        loop.run_until_complete(run())
        # run()が終了するまでここは通過しない
        _dt = self.scan_queue.get_nowait()             # キューバッファからデータを取り出す
        if _dt is None:
            self.main_win.SetStatusText('デバイスが見つかりません。')
        else:
            i = self.detect_list.GetFirstSelected()
            self.detect_list.SetItem(i, 0, 'スキャン完')  # 検出リストに接続状態を表示させる
            self.main_win.SetStatusText('デバイスのスキャンが完了しました。')
        return _dt

    def menu_service(self):
        """ 選択されたデバイスの全サービスを表示する処理 """
        _svdt = []

        async def run():
            try:
                _client = BleakClient(win_mon.DEVICE_ADDRESS)
            except bleak.exc.BleakError as _e:
                print(_e)
            else:
                try:
                    await _client.connect(timeout=20.0)  # 接続
                except bleak.exc.BleakError as _e:
                    print(_e)
                else:
                    if _client.is_connected:
                        # サーバーに登録されている全てのサービスを取得します。
                        _svcs = await _client.get_services()
                        for service in _svcs:
                            _svdt.append(service)
                        await _client.disconnect()  # 接続切断
                    else:
                        print('\nタイマーアウトみたいな')

        if win_mon.DEVICE_ADDRESS == "":
            self.main_win.SetStatusText('デバイスが選択されていません。')
            return
        self.main_win.SetStatusText('【サービス取得中】')
        # BLE通信タスクの起動を行う
        loop = asyncio.get_event_loop()
        loop.run_until_complete(run())
        # run()が終了するまでここは通過しない
        self.service_list.DeleteAllItems()       # サービスリストを空にする
        # サービスリストに追加する
        i = 0
        for l in _svdt:
            self.service_list.InsertItem(i, l.uuid)
            self.service_list.SetItem(i, 1, str(l.handle))
            self.service_list.SetItem(i, 2, l.description)
            i = i + 1
        if i == 0:
            self.main_win.SetStatusText('サービスは検出されませんでした(Detected again)。')
        else:
            self.main_win.SetStatusText(f'サービスを {i}件 検出しました。')
            i = self.detect_list.GetFirstSelected()
            self.detect_list.SetItem(i, 0, 'サービス完')  # 検出リストに接続状態を表示させる

    def menu_char(self):
        """ 選択されたUUIDの全キャラクタリスティックを表示する処理 """
        _svdt = []

        async def run():
            try:
                _client = BleakClient(win_mon.DEVICE_ADDRESS)
            except bleak.exc.BleakError as _e:
                print(_e)
            else:
                try:
                    await _client.connect(timeout=20.0)  # 接続
                except bleak.exc.BleakError as _e:
                    print(_e)
                else:
                    if _client.is_connected:
                        # サーバーに登録されている全てのサービスを取得します。
                        _svcs = _client.services
                        for service in _svcs:
                            # 指定のUUIDのみ処理する
                            if service.uuid == self.SERVICE_UUID:
                                _dt = service.characteristics
                                for char in _dt:
                                    # キャラクタリスティックの表示
                                    if "read" in char.properties:
                                        # "read"なら値を読み込む
                                        try:
                                            _value = bytes(await _client.read_gatt_char(char.uuid))
                                            _svdt.append(
                                                f"char: {char} ({','.join(char.properties)}), Value: {_value}"
                                            )
                                        except Exception as _e:
                                            _svdt.append(
                                                f"error: {char} ({','.join(char.properties)}), Value: {_e}"
                                            )
                                    else:
                                        _value = None
                                        _svdt.append(
                                            f"char: {char} ({','.join(char.properties)}), Value: {_value}"
                                        )
                                    # ディスクリプタの表示
                                    for descriptor in char.descriptors:
                                        try:
                                            _value = bytes(
                                                await _client.read_gatt_descriptor(descriptor.handle)
                                            )
                                            _svdt.append(f"┗desc: {descriptor}), Value: {_value}")
                                        except Exception as _e:
                                            _svdt.append(f"┗error: {descriptor}), Value: {_e}")
                        await _client.disconnect()  # 接続切断
                    else:
                        print('\nタイマーアウトみたいな')

        if self.SERVICE_UUID == "":
            self.main_win.SetStatusText('サービスUUIDが選択されていません。')
            return
        self.main_win.SetStatusText('【キャラクタリスティック取得中】')
        # BLE通信タスクの起動を行う
        loop = asyncio.get_event_loop()
        loop.run_until_complete(run())
        # run()が終了するまでここは通過しない
        self.char_list.DeleteAllItems()         # キャラクタリスティックリストを空にする
        # キャラクタリスティックリストに追加する
        i = 0
        for l in _svdt:
            s = l.split(':')
            self.char_list.InsertItem(i, s[0])  # "type"を取り出す
            self.char_list.SetItem(i, 2, s[2].replace(')', ''))        # "HANDLE"を取り出す
            self.char_list.SetItem(i, 3, s[4])  # "値"を取り出す
            self.char_list.SetItem(i, 4, s[3].replace(', Value', ''))  # "説明"を取り出す
            s = l.split()
            self.char_list.SetItem(i, 1, s[1])  # "UUID"を取り出す
            i = i + 1
        if i == 0:
            self.main_win.SetStatusText('キャラクタリスティックは検出されませんでした(Detected again)。')
        else:
            self.main_win.SetStatusText(f'キャラクタリスティックを {i}件 検出しました。')
            i = self.detect_list.GetFirstSelected()
            self.detect_list.SetItem(i, 0, 'キャラ完')  # 検出リストに接続状態を表示させる

    def menu_notify(self):
        """ 選択されたUUID情報から通知待ちを実行する処理 """
        def notification_handler(sender, data):
            """受信したデータを出力するシンプルな通知ハンドラー。"""
            _d = int.from_bytes(data, 'little')  # バイトオーダーは'little'とする
            print("{0}: {1} HEX({2}) 整数({3})".format(sender, data, data.hex(), _d))

        async def run_notify():
            try:
                _client = BleakClient(win_mon.DEVICE_ADDRESS)
            except bleak.exc.BleakError as _e:
                print(_e)
            else:
                try:
                    await _client.connect(timeout=20.0)  # 接続
                except bleak.exc.BleakError as _e:
                    print(_e)
                else:
                    if _client.is_connected:
                        # 通知受信の有効化、受信したらコールバック関数を処理します
                        try:
                            await _client.start_notify(win_mon.CHARACTERISTIC_UUID[0], notification_handler)
                        except bleak.exc.BleakError as _e:
                            print(f'[{win_mon.CHARACTERISTIC_UUID[0]}]このUUIDは通知では無い様です。')
                            print(_e)
                        else:
                            print(f'\nstart_notify:{self.NOTIFY_TIMER}秒間受信します。')
                            # 指定秒間(デフォルトは(10秒)受信したら通知受信を無効にする
                            await asyncio.sleep(self.NOTIFY_TIMER)
                            await _client.stop_notify(win_mon.CHARACTERISTIC_UUID[0])
                        await _client.disconnect()  # 接続切断
                    else:
                        print('\nタイマーアウトみたいな')

        kosu = len(win_mon.CHARACTERISTIC_UUID)
        if kosu == 0:
            self.main_win.SetStatusText('キャラクタリスティックUUIDが選択されていません。')
            return
        self.main_win.SetStatusText('【デバイスから通知中】')
        # BLE通信タスクの起動を行う
        loop = asyncio.get_event_loop()
        loop.run_until_complete(run_notify())
        # run_notify()が終了するまでここは通過しない
        self.main_win.SetStatusText(f'{self.NOTIFY_TIMER}秒間過ぎました。')

    def menu_read(self):
        """ 選択されたUUID情報から read を実行する処理 """
        async def run_read(kosu, kind, handle):
            try:
                _client = BleakClient(win_mon.DEVICE_ADDRESS)
            except bleak.exc.BleakError as _e:
                print(_e)
            else:
                try:
                    await _client.connect(timeout=20.0)  # 接続
                except bleak.exc.BleakError as _e:
                    print(_e)
                else:
                    if _client.is_connected:
                        print(f'start_read:{kosu}件readします。')
                        for _i in range(0, kosu):
                            if kind[_i] == 'char':
                                _value = bytes(await _client.read_gatt_char(win_mon.CHARACTERISTIC_UUID[_i]))
                            elif (kind[_i] == '┗desc') or (kind[_i] == 'desc'):
                                _value = bytes(await _client.read_gatt_descriptor(int(handle[_i])))
                            else:
                                _value = None
                            if _value is None:
                                print('Read できません。')
                            else:
                                _d = int.from_bytes(_value, 'little')  # バイトオーダーは'little'とする
                                print(f'Read Value: {_value} HEX({_value.hex()}) 整数({_d})')
                        await _client.disconnect()  # 接続切断
                    else:
                        print('\nタイマーアウトみたいな')

        kosu = len(win_mon.CHARACTERISTIC_UUID)
        if kosu == 0:
            self.main_win.SetStatusText('キャラクタリスティックUUIDが選択されていません。')
            return
        self.main_win.SetStatusText('【デバイスからread中】')
        # キャラクタリスティックリストから選択されている個数だけUUIDの種類とハンドルを得る
        kind = []
        handle = []
        for _i in range(0, kosu):
            _s = win_mon.CHARACTERISTIC_ATTR[_i].split(':')
            kind.append(_s[0])
            handle.append(_s[1])
        # BLE通信タスクの起動を行う
        loop = asyncio.get_event_loop()
        loop.run_until_complete(run_read(kosu, kind, handle))
        # run_notify()が終了するまでここは通過しない
        self.main_win.SetStatusText('')

    #---- サブ画面の表示関連関数 ----
    def menu_console(self):
        """ コンソール画面を表示する処理 """
        self.main_win.SetStatusText('')
        if self.TASK_CONSOLE_FLAG:
            self.main_win.SetStatusText('既に起動中です。')
            return
        self.TASK_CONSOLE_FLAG = True

        # 【コンソールフレームを作成】
        self.mon_win = wx.Frame(self.main_win, -1, 'コンソール画面',
                                size=(600, 300), style=wx.RESIZE_BORDER | wx.CLOSE_BOX | wx.CAPTION)
        self.mon_win.SetBackgroundColour("LIGHT BLUE")
        self.mon_win.Bind(wx.EVT_CLOSE, self.mon_winClose)  # 閉じる[X]のイベント
        self.CONSOLE_OBJ = win_mon(self.mon_win)            # コンソールフレームオブジェクト
        self.mon_win.SetIcon(self.icon)                     # アイコンの設定
        sys.stdout = self.CONSOLE_OBJ  # 標準出力をコンソールフレームオブジェクトに切り替える
        sys.stderr = self.CONSOLE_OBJ  # 標準エラー出力をコンソールフレームオブジェクトに切り替える

        # [画面クリアボタン作成]
        clear_btn = wx.Button(self.mon_win, wx.ID_ANY, '画面クリア', size=(80, 26))
        clear_btn.Bind(wx.EVT_LEFT_DOWN, self.click_clear_btn)

        # [レイアウトの設定]
        layout = wx.BoxSizer(wx.VERTICAL)
        layout.Add(clear_btn, 0, flag=wx.ALIGN_RIGHT)
        layout.Add(self.CONSOLE_OBJ.obj(), flag=wx.EXPAND | wx.ALL, border=10, proportion=1)
        self.mon_win.SetSizer(layout)

        # [コンソールフレームの表示]
        print('Hello!')
        self.mon_win.Show()

    def menu_setgatt(self):
        """ GATT設定情報画面を表示する処理 """
        self.main_win.SetStatusText('')
        if self.TASK_SETGATT_FLAG:
            self.main_win.SetStatusText('既に起動中です。')
            return
        self.TASK_SETGATT_FLAG = True

        # 【GATT設定情報フレームを作成】
        self.setgatt_win = wx.Frame(self.main_win, -1, 'GATT設定情報',
                                    size=(300, 350), style=wx.RESIZE_BORDER | wx.CLOSE_BOX | wx.CAPTION)
        self.setgatt_win.SetBackgroundColour("LIGHT BLUE")
        self.setgatt_win.Bind(wx.EVT_CLOSE, self.setgatt_winClose)  # 閉じる[X]のイベント
        self.setgatt_win.SetIcon(self.icon)                         # アイコンの設定

        # [メインパネルの作成]
        main_panel = wx.Panel(self.setgatt_win, wx.ID_ANY)

        # [ボタンサブパネル作成]
        btn_panel = wx.Panel(main_panel, wx.ID_ANY)
        # 記録帳を見るボタン
        self.look_btn = wx.Button(btn_panel, wx.ID_ANY, '記録帳を見る', size=(90, 26))
        self.look_btn.Bind(wx.EVT_LEFT_DOWN, self.click_look_btn)
        # 記録帳に登録ボタン
        self.record_btn = wx.Button(btn_panel, wx.ID_ANY, '記録帳に登録', size=(90, 26))
        self.record_btn.Bind(wx.EVT_LEFT_DOWN, self.click_record_btn)
        # ボタンレイアウトの設定
        btn_layout = wx.BoxSizer(wx.HORIZONTAL)
        btn_layout.Add(self.look_btn, 0, wx.ALL, 5)
        btn_layout.Add(self.record_btn, 0, wx.ALL, 5)
        btn_panel.SetSizer(btn_layout)

        # [GATT設定情報よこサブパネル作成]
        set_panel_yoko = wx.Panel(main_panel, wx.ID_ANY)
        # キャラクタリスティックのUUID用入力ボックス
        _s = ''
        _kosu = len(win_mon.CHARACTERISTIC_UUID)
        if _kosu != 0:
            _s = win_mon.CHARACTERISTIC_UUID[0]
            _s1 = 'キャラクタリスティックのUUID：' + f'[1/{_kosu}件です。]'
        else:
            _s1 = 'キャラクタリスティックのUUID：'+ '[0件です。]'
        self.uuid2_textctrl = wx.TextCtrl(set_panel_yoko, wx.ID_ANY, value=_s, style=wx.TE_PROCESS_ENTER, size=(250, 25))
        self.uuid2_textctrl.Bind(wx.EVT_TEXT_ENTER, self.uuid2_textctrl_enter)
        # スピンボタン
        self.spin_button = wx.SpinButton(set_panel_yoko, style=wx.SP_VERTICAL)
        self.spin_button.Bind(wx.EVT_SPIN, self.spin_up_down)
        # スピンボタンの値を設定する
        self.spin_button.SetMin(0)
        _kosu = len(win_mon.CHARACTERISTIC_UUID)
        if _kosu == 0:
            self.spin_button.SetMax(0)
        else :
            self.spin_button.SetMax(_kosu-1)

        # [GATT設定情報たてサブパネル作成]
        set_panel_tate = wx.Panel(main_panel, wx.ID_ANY)
        # デバイスのアドレス用入力ボックス
        adrs_come = wx.StaticText(set_panel_tate, wx.ID_ANY, 'デバイスのアドレス：')
        self.adrs_textctrl = wx.TextCtrl(set_panel_tate, wx.ID_ANY, value=win_mon.DEVICE_ADDRESS, style=wx.TE_PROCESS_ENTER)
        self.adrs_textctrl.Bind(wx.EVT_TEXT_ENTER, self.adrs_textctrl_enter)
        # サービスのUUID用入力ボックス
        uuid1_come = wx.StaticText(set_panel_tate, wx.ID_ANY, 'サービスのUUID：')
        self.uuid1_textctrl = wx.TextCtrl(set_panel_tate, wx.ID_ANY, value=self.SERVICE_UUID, style=wx.TE_PROCESS_ENTER)
        self.uuid1_textctrl.Bind(wx.EVT_TEXT_ENTER, self.uuid1_textctrl_enter)
        # 通知待ちの有効期間用入力ボックス
        notify_come = wx.StaticText(set_panel_tate, wx.ID_ANY, '通知待ちの有効期間(秒)：')
        notify_textctrl = wx.TextCtrl(set_panel_tate, wx.ID_ANY, value=str(self.NOTIFY_TIMER), style=wx.TE_PROCESS_ENTER)
        notify_textctrl.Bind(wx.EVT_TEXT_ENTER, self.notify_textctrl_enter)
        # 注意のコメント
        note_come = wx.StaticText(set_panel_tate, wx.ID_ANY, '注意）データを入力した時は必ず[Enter]を押して下さい。\n\r')
        note_come.SetForegroundColour('#EE0000')
        # 操作のコメント
        op_come = wx.StaticText(set_panel_tate, wx.ID_ANY, '( \"type;UUID;ハンドル\" の構成で入力します。)')
        # キャラクタリスティックのタイトル
        self.uuid2_come = wx.StaticText(set_panel_tate, wx.ID_ANY, _s1)

        # GATT設定情報たてレイアウトの設定
        tate_layout = wx.BoxSizer(wx.VERTICAL)
        tate_layout.Add(note_come, 0)
        tate_layout.Add(notify_come, 0)
        tate_layout.Add(notify_textctrl, flag=wx.EXPAND | wx.ALL, border=5)
        tate_layout.Add(adrs_come, 0)
        tate_layout.Add(self.adrs_textctrl, flag=wx.EXPAND | wx.ALL, border=5)
        tate_layout.Add(uuid1_come, 0)
        tate_layout.Add(self.uuid1_textctrl, flag=wx.EXPAND | wx.ALL, border=5)
        tate_layout.Add(self.uuid2_come, 0)
        tate_layout.Add(op_come, 0)
        set_panel_tate.SetSizer(tate_layout)

        # GATT設定情報よこレイアウトの設定
        yoko_layout = wx.BoxSizer(wx.HORIZONTAL)
        yoko_layout.Add(self.uuid2_textctrl, 0, flag=wx.ALL, border=5)
        yoko_layout.Add(self.spin_button, 0, flag=wx.EXPAND)
        set_panel_yoko.SetSizer(yoko_layout)

        # 【メインパネルにサブパネルを縦に並べる】
        main_layout = wx.BoxSizer(wx.VERTICAL)
        main_layout.Add(btn_panel, 0)
        main_layout.Add(set_panel_tate, 0, flag=wx.EXPAND | wx.ALL, border=5)
        main_layout.Add(set_panel_yoko, 0, flag=wx.EXPAND | wx.ALL, border=5)
        main_panel.SetSizer(main_layout)

        # [GATT設定情報の表示]
        self.setgatt_win.Show()

    ###############################################################################
    # Windows画面のメイン
    ###############################################################################
    def main(self):
        # ２重起動チェック
        instance_name = u"%s-%s" % (app.GetAppName(), wx.GetUserId())
        instance = wx.SingleInstanceChecker(instance_name)
        if instance.IsAnotherRunning():
            wx.Exit()
        # 【メインウインドウを作成】
        # 基本フレームの生成
        self.main_win = wx.Frame(None, wx.ID_ANY, 'skＢＬＥツール V1.10', size=(800, 700))
        self.main_win.Centre()  # ウインドウの表示位置は中央
        self.main_win.Bind(wx.EVT_CLOSE, self.FrameClose)  # 閉じる[X]のイベント
        self.icon = wx.Icon(r"skBLEtool.ico", wx.BITMAP_TYPE_ICO)
        self.main_win.SetIcon(self.icon)                   # アイコンの設定
        # ステータスバーの作成
        self.main_win.CreateStatusBar()
        self.main_win.SetStatusText('zattouka.net')
        # 【メニューを作成】
        # 「ファイル」メニュー作成
        File_menu = wx.Menu()
        File_menu.Append(wx.ID_EXIT, "終了(&X)")
        # 「デバイス」メニュー作成
        device_menu = wx.Menu()
        device_menu.Append(1, "検出(&K)")
        # ”接続”サブメニュー作成
        Connect_sub_menu = wx.Menu()
        Connect_sub_menu.Append(3, "サービス(S)\tCtrl+S")
        Connect_sub_menu.Append(4, "Characteristics(C)\tCtrl+C")
        Connect_sub_menu.Append(5, "通知(N)\tCtrl+N")
        Connect_sub_menu.Append(8, "read(R)\tCtrl+R")
        device_menu.AppendSubMenu(Connect_sub_menu, '接続')
        device_menu.AppendSeparator()
        device_menu.Append(2, "スキャン(&S)")
        device_menu.Append(9, "オブザーバー(&O)")

        # 「機能」メニュー作成
        function_menu = wx.Menu()
        # 「Scan + read'」サブメニュー作成
        function_sub_menu = wx.Menu()
        function_sub_menu.Append(10, "開始")
        function_sub_menu.Append(11, "停止")
        function_menu.AppendSubMenu(function_sub_menu, 'Scan + read')

        # 「表示」メニュー作成
        display_menu = wx.Menu()
        display_menu.Append(6, "コンソール(&C)")
        display_menu.Append(7, "GATT設定情報(&I)")
        # 「ヘルプ」メニュー作成
        help_menu = wx.Menu()
        help_menu.Append(wx.ID_HELP, "マニュアル(&M)")

        # メニューバーの生成
        menu_bar = wx.MenuBar()
        # メニューをメニューバーへ登録
        menu_bar.Append(File_menu, 'ファイル(&F)')
        menu_bar.Append(device_menu, 'デバイス(&D)')
        menu_bar.Append(display_menu, '表示(&V)')
        menu_bar.Append(function_menu, '機能(&K)')
        menu_bar.Append(help_menu, 'ヘルプ(&H)')
        # フレームにメニュバー登録
        self.main_win.SetMenuBar(menu_bar)
        self.main_win.Bind(wx.EVT_MENU, self.SelectMenu)

        # 【最上位パネルの作成】
        main_top_panel = wx.Panel(self.main_win, wx.ID_ANY)
        main_top_panel.SetBackgroundColour("LIGHT BLUE")

        # 【検出リストコントロールの作成】
        self.detect_list = wx.ListCtrl(main_top_panel, wx.ID_ANY,
                                       style=wx.LC_AUTOARRANGE | wx.LC_REPORT, size=(800, 150))
        # カラム
        self.detect_list.InsertColumn(0, "操作", wx.LC_ALIGN_LEFT, 60)
        self.detect_list.InsertColumn(1, "アドレス", wx.LIST_FORMAT_CENTRE, 110)
        self.detect_list.InsertColumn(2, "名前", wx.LIST_FORMAT_CENTRE, 100)
        self.detect_list.InsertColumn(3, "UUID", wx.LIST_FORMAT_CENTRE, 240)
        self.detect_list.InsertColumn(4, "RSSI", wx.LIST_FORMAT_CENTRE, 40)
        self.detect_list.InsertColumn(5, "詳細(details)", wx.LIST_FORMAT_CENTRE, 120)
        self.detect_list.InsertColumn(6, "製造元のデータ", wx.LIST_FORMAT_CENTRE, 130)
        # イベントの登録
        self.detect_list.Bind(wx.EVT_LIST_ITEM_SELECTED, self.detect_list_select)
        self.detect_list.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.detect_list_deselect)

        # 【サービスリストコントロールの作成】
        wx.StaticText(main_top_panel, wx.ID_ANY, 'サービスのリスト：', pos=(10, 172))
        self.service_list = wx.ListCtrl(main_top_panel, wx.ID_ANY,
                                        style=wx.LC_AUTOARRANGE | wx.LC_REPORT, size=(800, 200))
        # カラム
        self.service_list.InsertColumn(0, "UUID", wx.LC_ALIGN_LEFT, 300)
        self.service_list.InsertColumn(1, "handle", wx.LC_ALIGN_LEFT, 50)
        self.service_list.InsertColumn(2, "説　明 ", wx.LC_ALIGN_LEFT, 450)
        # イベントの登録
        self.service_list.Bind(wx.EVT_LIST_ITEM_SELECTED, self.service_list_select)
        self.service_list.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.service_list_deselect)

        # 【キャラクタリスティックリストコントロールの作成】
        wx.StaticText(main_top_panel, wx.ID_ANY, 'キャラクタリスティック／ディスクリプタのリスト：', pos=(10, 392))
        self.char_list = wx.ListCtrl(main_top_panel, wx.ID_ANY,
                                     style=wx.LC_AUTOARRANGE | wx.LC_REPORT, size=(800, 200))
        # カラム
        self.char_list.InsertColumn(0, "type", wx.LC_ALIGN_LEFT, 50)
        self.char_list.InsertColumn(1, "UUID", wx.LC_ALIGN_LEFT, 250)
        self.char_list.InsertColumn(2, "handle", wx.LC_ALIGN_LEFT, 50)
        self.char_list.InsertColumn(3, "説　明 ", wx.LC_ALIGN_LEFT, 350)
        self.char_list.InsertColumn(3, "値 ", wx.LC_ALIGN_LEFT, 100)
        # イベントの登録
        self.char_list.Bind(wx.EVT_LIST_ITEM_SELECTED, self.char_list_select)
        self.char_list.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.char_list_deselect)

        # ［レイアウトの設定］
        layout1 = wx.BoxSizer(wx.VERTICAL)
        layout1.Add(self.detect_list, flag=wx.EXPAND | wx.ALL, border=10, proportion=1)
        layout1.Add(self.service_list, flag=wx.EXPAND | wx.ALL, border=10, proportion=1)
        layout1.Add(self.char_list, flag=wx.EXPAND | wx.ALL, border=10, proportion=1)
        main_top_panel.SetSizer(layout1)

        # 【タイマーの作成】
        self.timer = wx.Timer()
        self.timer.Bind(wx.EVT_TIMER, self.OnTimer)

        # 【メインウインドウの表示】
        self.main_win.Show()
# クラスwin_appの終わり

if __name__ == "__main__":
    app = wx.App()  # アプリケーションクラスの生成
    win = win_app()
    win.main()
    app.MainLoop()  # 【ウインドウを動かす(イベントの発生待ち)】
